/*
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2009, 2015.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 */
(function() {
dojo.provide("com.ibm.rdm.web.calm.CalmLinkUtils");

dojo.require("dojox.xml.parser");
dojo.require("dojo.DeferredList");
dojo.require("dojo.i18n");

dojo.require("com.ibm.rdm.web.artifactgrid.ArtifactConstants");
dojo.require("com.ibm.rdm.web.calm.CalmConstants");
dojo.require("com.ibm.rdm.web.calm.OSLCV1LinkUtils");
dojo.require("com.ibm.rdm.web.calm.OSLCV2LinkUtils");
dojo.require("com.ibm.rdm.web.common.MessageHandler");
dojo.require("com.ibm.rdm.web.common.MimeTypes");
dojo.require("com.ibm.rdm.web.links.Link");
dojo.require("com.ibm.rdm.web.links.LinkTypeUtils");
dojo.require("com.ibm.rdm.web.model.resources.ResourceManager");
dojo.require("com.ibm.rdm.web.parsers.xml2json");
dojo.require("com.ibm.rdm.web.query.SavedFilterUtils");
dojo.require("com.ibm.rdm.web.server.Headers");
dojo.require("com.ibm.rdm.web.server.client.RDMClient");
dojo.require("com.ibm.rdm.web.server.client.utils.LinkUtils");
dojo.require("com.ibm.rdm.web.util.queryutils");
dojo.require("com.ibm.rdm.web.common.StatusDisplay");
dojo.require("com.ibm.rdm.web.common.ProjectCache");

dojo.requireLocalization("com.ibm.rdm.web.calm", "CalmLinkUtils");
dojo.requireLocalization("com.ibm.rdm.web.links", "LinksListWidget");

var ArtifactConstants = com.ibm.rdm.web.artifactgrid.ArtifactConstants;
var CalmConstants = com.ibm.rdm.web.calm.CalmConstants;
var queryutils    = com.ibm.rdm.web.util.queryutils;
var RDMClient     = com.ibm.rdm.web.server.client.RDMClient;
var Headers		  = com.ibm.rdm.web.server.Headers;
var LinkUtils     = com.ibm.rdm.web.server.client.utils.LinkUtils;
var LinkTypeUtils = com.ibm.rdm.web.links.LinkTypeUtils;
var MimeTypes     = com.ibm.rdm.web.common.MimeTypes;
var StatusDisplay = com.ibm.rdm.web.common.StatusDisplay;
var MessageHandler = com.ibm.rdm.web.common.MessageHandler;
var ProjectCache = com.ibm.rdm.web.common.ProjectCache;

var messages = dojo.i18n.getLocalization("com.ibm.rdm.web.calm", "CalmLinkUtils");
var calmMessages = dojo.i18n.getLocalization("com.ibm.rdm.web.links", "LinksListWidget");

var _XML_HEADER  = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"; 
var _XML_DOCTYPE = "<!DOCTYPE rdf:RDF [<!ENTITY xsd \"" + queryutils.XML_SCHEMA + "\">]>\n";

var XML_NAMESPACES = { 
	"xmlns:rrm":  queryutils.RRM_NAMESPACE,
	"xmlns:rdf":  queryutils.RDF_NAMESPACE,
	"xmlns:xsd":  queryutils.XML_SCHEMA,
	"xmlns:rdfs": queryutils.RDFS_NAMESPACE
};

var UPDATE_CACHE_EVENT = {
		CREATE: "create",
		DELETE: "delete"
};


//Provides utility methods for processing Calm links
com.ibm.rdm.web.calm.CalmLinkUtils = {

	backLinkDeferred :null,

	
	createLinks: function(
		projectAreaLink,
		calmLinks,
		/*Boolean?*/doFailFast,
		resourceInfos,
		/*Boolean?*/doShowIntermediateStatusMessages)
	{
		var fakeDeferred =  new dojo.Deferred();
		var failFastHasOccurred = false,
			i = 0,
			numLinks = calmLinks.length;
		var createLinksSynchronously = dojo.hitch(this, function(){
			// Note: This function creates links one by one in a synchronous fashion.
			// We do this due to the fact that for every new "front link" we create we
			// ALSO create a "back link".  However, this back link may be a single
			// resource that's updated to account for >1 front link.  For example,
			// when creating an "Implemented By" link from two requirements to a work
			// item, we create two "front links" and do two PUTs on the same resource
			// to represent the back links (in OSLCV2LinkUtils's createBackLink method).
			// In other words, we create the links in a synchronous fashion to enable
			// us to send off the correct ETag when we make these consecutive PUTs
			// (for the back links).
			if (i === numLinks) {
				if (doShowIntermediateStatusMessages) {
					StatusDisplay.hideStatusMessage();
				}
				fakeDeferred.callback(true);
			} else {
				if (doShowIntermediateStatusMessages) {
					StatusDisplay.showStatusMessage(
						dojo.string.substitute(
							messages["savingLinkNofM"],
							[i+1, numLinks]
						)
					);
				}
				var createLinkDeferred =
					this.createLink(
						projectAreaLink,
						calmLinks[i],
						resourceInfos ? [resourceInfos[i]] : null,
						(numLinks > 1 ? true : false)
					);
				createLinkDeferred.then(
					function(){
						i++;
						createLinksSynchronously();
					},
					function(response){
						i++;
						if (doFailFast) {
							if (doShowIntermediateStatusMessages) {
								StatusDisplay.hideStatusMessage();
							}
							fakeDeferred.errback(response);
						} else {
							var responseText = response.responseText;
							// Note: We're ignoring RRRS4506W ("The link was not created because an
							// identical link already exists for the artifact.") ... but any other
							// error should be reported.
							if (responseText && responseText.indexOf("CRRRS4506W") === -1) {
								MessageHandler.showXhrError(response, response.ioArgs);
							}
							createLinksSynchronously();
						}
					}
				);
			}
		});
		createLinksSynchronously();
		
		return fakeDeferred;
	},
	
		
	createLink: function(projectAreaUri, calmLink, resourceInfos, /*Boolean?*/skipErrors) {
		var createLinkDeferred = new dojo.Deferred();
		
		var innerCreateLinks = dojo.hitch(this, function(isCMEnabled) {
			
			/*
			 * The create link process follows 1 of 3 workflows depending on whether it's RM to RM or CM enabled.  
			 * The RM to RM link type has priority over all other workflow options regardless of CM status.
			 * 
			 * RM to RM:
			 *   1) Create the local (front) link
			 * 
			 * CM Enabled (opt-in):
			 *   1) Create the remote owned target -> source (back) link
			 *      - only the derived remote (back) link is created since links are owned by the remote resources.
			 *      - check for link on remote resource to avoid creating duplicates
			 *   2) Immediately update the link indexer rather then wait for the background process
			 *   
			 * CM Disabled (opt-out):
			 *   1) Create the remote owned target -> source (back) link first
			 *   2) Create the local (front) link second
			 *      - if the create local (front) link request fails then delete the remote (back) link
			 */
			var continueLinking = dojo.hitch(this, function() {
				
				var linkTypeURLToCalmLinkType = this._linkTypeURLToCalmLinkType(projectAreaUri, calmLink.getLinkType());
				linkTypeURLToCalmLinkType.addCallback(this, function(linkType) {
			
					if(LinkTypeUtils.isRMtoRMLinkType(linkType)) {
						// RM to RM
						this._createLocalResourceOwnedLinkForRMtoRMHandler(createLinkDeferred, projectAreaUri, calmLink, skipErrors);
					}
					else if (isCMEnabled) {
						// Opt-in
						this._createRemoteResourceOwnedLinksWhenCmEnabled(createLinkDeferred, projectAreaUri, calmLink, linkType, skipErrors);
					}
					else {
						// Opt-out
						this._createRemoteAndLocalResourceOwnedLinksWhenCmDisabled(createLinkDeferred, projectAreaUri, calmLink, skipErrors);
					}
					
					return createLinkDeferred;
				});
			});

			if (resourceInfos && resourceInfos[0]){
				if ((typeof resourceInfos[0].isStub === "function") && resourceInfos[0].isStub()){
					var cleanupLink = dojo.hitch(this, function(){
						if (calmLink && calmLink._subject === resourceInfos[0].stubId){
							calmLink._subject = resourceInfos[0].getResourceURI();
						}
						else if (calmLink && calmLink._object === resourceInfos[0].stubId){
							calmLink._object = resourceInfos[0].getResourceURI();
						}
					});
					resourceInfos[0].addPostSaveOperation(cleanupLink);
					resourceInfos[0].addPostSaveOperation(continueLinking);
					var linked = new dojo.Deferred();
					linked.callback(true);
					return linked;
				}
				else{
					return continueLinking();
				}
			}
			else{
				return continueLinking();
			}
		});

		// get the CM state and then invoke the function to create links
		ProjectCache.getCmStateDeferred(projectAreaUri)
			.then(innerCreateLinks);
	
		return createLinkDeferred;
	},
	
	_createRemoteResourceOwnedLinksWhenCmEnabled: function(createLinkDeferred, projectAreaUri, calmLink, linkType, /*optional*/skipErrors) {
		// summary:
		//		Create the remote back link and update the local cache when the project Global CM state is enabled (opt-in)
		// description:
		//		Attempt to create the remote (back) link on the remote resource.
		//		Update the local cache if required.
		// createLinkDeferred:
		//		Top level deferred object with call back functions
		// projectAreaUri:
		//		The URI of the project to update.
		// calmLink:
		//		A link object to be updated including source and target details.
		// linkType:
		//		The link type uri.
		// skipErrors: 
		//		Optional argument.  If true, xhr errors that occur will be ignored and will not be displayed in the application
		//		messageBox.  The value of this parameter effects how the callback response data is returned.
		// returns:
		//		Deferred.  When skipErrors is true, the callback response will be a deferred response,
		//		which will include both successful response and error response.  When skipErrors is false or undefined,
		//		the callback response will be set to true or false.		

		// GLobal CM enabled (opt-in):
		// Only the derived remote (back) link is created since links are owned by the remote resources.
		
		// 1) create the remote owned target -> source (back) link first				
		//      - check for link on remote resource to avoid creating duplicates
		var checkForExistingLinkDeferred = this._checkLinkInCMLinkIndexer(projectAreaUri, calmLink.getSubject(), calmLink.getObject(), calmLink.getLinkType(), skipErrors);
		checkForExistingLinkDeferred.addCallback(this, function(response) {
			
			// 200 indicates a link does not already exist and it is ok to continue creating the new link
			if (response.ioArgs.xhr.status != "200") {
				createLinkDeferred.callback(false);
				return;
			}
		
			var remoteResourceOwnedLinkDeferred = this._createRemoteResourceOwnedLinkDeferred(projectAreaUri, calmLink);

			remoteResourceOwnedLinkDeferred.addCallback(this, function(response) {
				// 2) If Global CM is enabled (opt-in) then:
				//      - update the cache 
				if (linkType !== LinkTypeUtils.SYSTEM_LINK_TYPE_TRACKS_CHANGE_SET) {
					// Check CM status to determine whether the cache should be updated
					// Tracks Change Set do not use the JAF LDX cache
					// Legacy check because artifacts should not be linking to Tracks Change Set
					var updateCache = this._updateLinkInCMLinkIndexer(projectAreaUri, calmLink.getSubject(), calmLink.getObject(), calmLink.getLinkType(), UPDATE_CACHE_EVENT.CREATE);
					updateCache.then(function(response) {
						// assuming if the cache update fails, it will eventually be resolved by the LDX update process (1 minute)
						// Should we notify the user that it may take a few minutes to update on failure?
						createLinkDeferred.callback(true);
					});
				} else {
					createLinkDeferred.callback(true);
				}
			});
			
			remoteResourceOwnedLinkDeferred.addErrback(this, this._processCreateLinkError, createLinkDeferred, skipErrors, 7253); // delete failed too						
		});			
	},
			
	_createRemoteAndLocalResourceOwnedLinksWhenCmDisabled: function(createLinkDeferred, projectAreaUri, calmLink, /*optional*/skipErrors) {
		// summary:
		//		Create the remote back link and local front link when the project Global CM is disabled (opt-out) 
		// description:
		//		Attempt to create the remote (back) link on the remote resource.
		//		First check if user has permission to create local links
		//		Only if the remote (back) link is created and do we then attempt to create the local (front) link.
		// createLinkDeferred:
		//		Top level deferred object with call back functions
		// projectAreaUri:
		//		The URI of the project to update.
		// calmLink:
		//		A link object to be updated including source and target details.
		// skipErrors: 
		//		Optional argument.  If true, xhr errors that occur will be ignored and will not be displayed in the application
		//		messageBox.  The value of this parameter effects how the callback response data is returned.
		// returns:
		//		Deferred.  When skipErrors is true, the callback response will be a deferred response,
		//		which will include both successful response and error response.  When skipErrors is false or undefined,
		//		the callback response will be set to true or false.
				
		// Global CM disabled (opt-out):
		// The order in which both local (front) and remote (back) links are created does not matter.
		// Since creating the remote (back) link has a higher risk it is created first to minimise scenarios
		// where a local (front) link exists without the corresponding remote (back) link.
		
		// 1) check permission to create local links
		// 2) create the remote owned target -> source (back) link first
		// 3) create the local -> remote (front) link second
		var remoteAndlocalResourceOwnedLinkDeferred = this._verifyCreateLinkPermission(projectAreaUri)
				.then(dojo.hitch(this, this._createRemoteResourceOwnedLinkDeferred, projectAreaUri, calmLink))
				.then(dojo.hitch(this, this._createLocalResourceOwnedLinkHandler, createLinkDeferred, projectAreaUri, calmLink, skipErrors))
				.otherwise(dojo.hitch(this, this._processCreateLinkError, createLinkDeferred, skipErrors, 7255));
		
		return remoteAndlocalResourceOwnedLinkDeferred;
	},
			
	_createLocalResourceOwnedLinkForRMtoRMHandler: function(createLinkDeferred, projectAreaUri, calmLink, /*optional*/skipErrors) {
		// summary:
		//		Set the deferred object that creates the local front link for RM to RM link types
		// createLinkDeferred:
		//		Top level deferred object with call back functions
		// projectAreaUri:
		//		The URI of the project to update.
		// calmLink:
		//		A link object containing the source and target details.
		// skipErrors: 
		//		Optional argument.  If true, xhr errors that occur will be ignored and will not be displayed in the application
		//		messageBox.  The value of this parameter effects how the callback response data is returned.
		// returns:
		//		Deferred.  When skipErrors is true, the callback response will be a deferred response,
		//		which will include both successful response and error response.  When skipErrors is false or undefined,
		//		the callback response will be set to true or false.
				
		// 1) check permission to create local links
		// 2) create the local -> remote (front) link 
		var localResourceOwnedLinkDeferred = this._verifyCreateLinkPermission(projectAreaUri)
				.then(dojo.hitch(this, this._createLocalResourceOwnedLink, projectAreaUri, calmLink, skipErrors))
				.then(dojo.hitch(this, this._postCreateLink, createLinkDeferred))
				.otherwise(dojo.hitch(this, this._processCreateLinkError, createLinkDeferred, skipErrors, 7255));
		
		return localResourceOwnedLinkDeferred;
	},
		
	_createLocalResourceOwnedLinkHandler: function(createLinkDeferred, projectAreaUri, calmLink, /*optional*/skipErrors) {
		// summary:
		//		Execute the deferred object that creates the local front link
		// createLinkDeferred:
		//		Top level deferred object with call back functions
		// projectAreaUri:
		//		The URI of the project to update.
		// calmLink:
		//		A link object containing the source and target details.
		// skipErrors: 
		//		Optional argument.  If true, xhr errors that occur will be ignored and will not be displayed in the application
		//		messageBox.  The value of this parameter effects how the callback response data is returned.
		// returns:
		//		Deferred.  When skipErrors is true, the callback response will be a deferred response,
		//		which will include both successful response and error response.  When skipErrors is false or undefined,
		//		the callback response will be set to true or false.
		
		var localResourceOwnedLinkDeferred = this._createLocalResourceOwnedLink(projectAreaUri, calmLink, skipErrors)
		    	.then(dojo.hitch(this, this._postCreateLink, createLinkDeferred))
		    	.otherwise(dojo.hitch(this, this._processCreateLocalResourceOwnedLinkError, createLinkDeferred, projectAreaUri, calmLink, skipErrors));
				
		return localResourceOwnedLinkDeferred;
	},
		
	_createLocalResourceOwnedLink:function(
			projectAreaUri,
			/*com.ibm.rdm.web.links.Link*/calmLink,
			/*Boolean?*/skipErrors) {
		// summary:
		//		Create the local owned resource (front) link
		// projectAreaUri:
		//		The URI of the project to update.
		// calmLink:
		//		A link object containing the source and target details.
		// skipErrors: 
		//		Optional argument.  If true, xhr errors that occur will be ignored and will not be displayed in the application
		//		messageBox.  The value of this parameter effects how the callback response data is returned.
		// returns:
		//		Deferred.  When skipErrors is true, the callback response will be a deferred response,
		//		which will include both successful response and error response.  When skipErrors is false or undefined,
		//		the callback response will be set to true or false.			
		
		var skipNotificationForOptInProjects = false;
		ProjectCache.getCmStateDeferred(projectAreaUri).then(dojo.hitch(this, function(enabled) {
			skipNotificationForOptInProjects = enabled;
		}));
		
		return LinkUtils.saveLink(
			projectAreaUri,
			calmLink,
			skipNotificationForOptInProjects,				// skipNotification
			skipErrors
		);
	},
	
	_updateLinkInCMLinkIndexer: function(projectURI, source, target, type, action, skipErrors) {
		// summary:
		//		Used for opt-in projects to add/remove a CALM link to the local cache (ie Link Index Provider)
		//		The action var determines whether the link is added or removed from the link indexer.
		// 
		var deferred = new dojo.Deferred();
		if ( target && target.indexOf("?oslc.configuration") > -1 ) {
			// remove that
			target = target.substring(0, target.indexOf("?oslc.configuration"));
		}
		
		var uriWithUpdate = com.ibm.rdm.web.FRONTING_SERVICE_LINKS_CACHE + "?sourceURL=" + encodeURIComponent(source) + "&targetURL=" + encodeURIComponent(target) + "&typeURL=" + encodeURIComponent(type) + "&action=" + action;
		
		var errorHandler = function(response) { 
			deferred.errback(response);
			return response;
		};
		
		var updateLinkCache = function() { 
			var saveDeferred = RDMClient.doPost({
				uri: uriWithUpdate,
				projectURI: projectURI,
				skipErrors: skipErrors,
				xhrArgs: {
					headers: {
						"Content-Type": MimeTypes.CONTENT_TYPE_RDF_XML
					}
				}
			});
			saveDeferred.addCallback(function(response) {
				deferred.callback(response);
				return response;
			});
			saveDeferred.addErrback(errorHandler);
		};
		
		
		// first check if the link exists in the cache already.
		updateLinkCache();

		return deferred;
	},
	
	_checkLinkInCMLinkIndexer: function(projectURI, source, target, type, skipErrors) {
		// summary:
		//		This checks to see if the same CALM link already exists in the Link Indexer for opt-in projects.  
		// description:
		//		When a 200 is returned, it should be fine to create the CALM link.  Otherwise, 
		//		we should inform the user that this is a duplicate by returning 403 - forbidden.
		//
		// returns: 200 if the link does not exist else 403 (forbidden)
		// 
		var checkExistsURI = com.ibm.rdm.web.FRONTING_SERVICE_LINKS_CACHE 
			+ "?sourceURL=" + encodeURIComponent(source) 
			+ "&targetURL=" + encodeURIComponent(target) 
			+ "&typeURL=" + encodeURIComponent(type)
			+ "&preventCache=" + Date.now();
		var checkDeferred = RDMClient.doGet({
			uri: checkExistsURI,
			projectURI: projectURI,
			skipErrors: skipErrors
		});
		return checkDeferred;
	},
		
	_verifyCreateLinkPermission: function(projectAreaUri) {
		// summary:
		//		Verify the user has permission to create CLM links. 
		// description:
		//		If the user does not have permission to create CLM links then
		//		throw an error with the relevant error code. 
		// projectAreaUri:
		//		The URI of the project to update.
		// returns:
		//		Deferred.  The deferred object that throws the potential error.

		var projectOperationsDeferred = com.ibm.rdm.web.common.ProjectCache.getProjectPermissions(projectAreaUri);
		var checkPermissions = projectOperationsDeferred
			.then(function(projectOperations) {
					var isCreateLinkAllowed = projectOperations.item(com.ibm.rdm.web.common.ProjectCache.MODIFY_CLM_LINK);
					if (!isCreateLinkAllowed) {
						throw new Error(7320); // permission denied message id
					}
			});
		
		return checkPermissions;
	},

	_postCreateLink: function(createLinkDeferred) {
		// summary:
		//		Generic post processing function to notify link creation process to continue with next step
		// createLinkDeferred:
		//		Top level deferred object with call back functions
		createLinkDeferred.callback(true);
	},	
	
	_processCreateLocalResourceOwnedLinkError: function(createLinkDeferred, projectAreaUri, calmLink, skipErrors, response) {
		// summary:
		//		Process the error encountered when creating a local link. 
		//		Called when an error has already occurred and is a best attempt to delete remote back link
		// createLinkDeferred:
		//		Top level deferred object with call back functions		
		// projectAreaUri:
		//		The URI of the project to update.
		// calmLink:
		//		A link object containing the source and target details
		// skipErrors: 
		//		The value of this parameter effects how the callback response data is returned.
		// response:
		//		The xhr response
		
		if (response && response.ioArgs && response.ioArgs.xhr.responseText.indexOf("CRRRS4506W") >= 0) {
			// Link cannot be created because an identical link already exists.
			// Skip attempt to delete remote link  
			return this._processCreateLinkError(createLinkDeferred, skipErrors, 7653, response);
		}
		else {
			return this.deleteBackLink(projectAreaUri, calmLink)
							.then(dojo.hitch(this, this._processCreateLinkError, createLinkDeferred, skipErrors, 7255, response)) // original error
							.otherwise(dojo.hitch(this, this._processCreateLinkError, createLinkDeferred, skipErrors, 7254, response)); // delete failed too	
		}		
	},
	
	_processCreateLinkError: function(createLinkDeferred, skipErrors, defaultErrorId, response) {
		// summary:
		//		Generic function to perform the error handling when an exception is thrown while creating a link
		//		and a response exists with additional information.
		// description:
		//		Process the error by displaying a warning message with the error id and
		//		invoking the top level callback to skip the remaining create link processing.
		// createLinkDeferred:
		//		Top level deferred object with call back functions
		// skipErrors: 
		//		Optional argument.  If true, xhr errors that occur will be ignored and will not be displayed in the application
		//		messageBox.  The value of this parameter effects how the callback response data is returned.
		// defaultErrorId:
		//		The IBM message box error id which may be overridden depending on the response
		// response:
		//		The xhr response with optional details for updating the error id
		//		Best effort attempt and subject to response object changes
		
		if (skipErrors) {
			createLinkDeferred.errback(response);
		} else {
			var errorId = defaultErrorId;
			
			// check for thrown errors with error id
			if (response) {
				if (!isNaN(response.message)) {
					errorId = response.message;
				}
				else if (response.ioArgs) {
					if (response.ioArgs.xhr.responseText.indexOf("CRRRS4506W") >= 0) {
						errorId = 7653; // local - identical link already exists 
					}
					// the following are a best effort attempt to parse known remote server exceptions
					else if (response.ioArgs.xhr.status === 403) {
						errorId = 7259; // remote - forbidden
					}
					else if (response.ioArgs.xhr.responseText.indexOf("Error 403: Forbidden") >= 0) {
						errorId = 7259; // DNG: remote - forbidden
					}
					else if (response.ioArgs.xhr.responseText.indexOf("<oslc:statusCode>403</oslc:statusCode>") >= 0) {
						errorId = 7259; // QM: remote - forbidden
					}
				}
			}
			
			this._putUpWarning(errorId); // can not create link
			createLinkDeferred.callback(false); // Error creating the link, return false
		}
	},	
	
	_linkTypeURLToCalmLinkType: function(projectAreaUrl, linkTypeURL) {
		var fakeDeferred = new dojo.Deferred();	
		if ( linkTypeURL === LinkTypeUtils.SYSTEM_LINK_TYPE_TRACKS_CHANGE_SET ) {
			fakeDeferred.callback(linkTypeURL);
		} else {
			var linkTypesDeferred = LinkTypeUtils.getLinkTypesForProject(projectAreaUrl);
			linkTypesDeferred.addCallback(this, function(linktypes) {
				for (var i = 0; i < linktypes.length; i++) {	
					if (linktypes[i].getResourceURI() == linkTypeURL) {
						 fakeDeferred.callback(linktypes[i].getSameAs());
						 break;
					}
				}
			});
		}
		return fakeDeferred;
	},
	
	_getAcceptHeaderForLinkType: function(linkTypeSameAs, isCollection) {
		var acceptHeader = null;
		switch (linkTypeSameAs) {
			case LinkTypeUtils.SYSTEM_LINK_TYPE_IMPLEMENTED_BY:
			case LinkTypeUtils.SYSTEM_LINK_TYPE_TRACKED_BY:
			case LinkTypeUtils.SYSTEM_LINK_TYPE_AFFECTED_BY:
				acceptHeader = CalmConstants.OSLC_CM.OSLC_CM_CHANGE_REQUEST_HEADER;
				break;
			case LinkTypeUtils.SYSTEM_LINK_TYPE_VALIDATED_BY:
				acceptHeader = isCollection ? CalmConstants.OSLC_QM.OSLC_QM_TEST_PLAN_HEADER : CalmConstants.OSLC_QM.OSLC_QM_TEST_CASE_HEADER;
				break;
		}
		return acceptHeader;
	},

	_createRemoteResourceOwnedLinkDeferred: function(projectAreaUri, calmLink, /*optional*/ retryDeferred) {
		// summary:
		//		Get the deferred object that creates the remote (back) link
		// projectAreaUri:
		//		The URI of the project to update.
		// calmLink:
		//		A link object containing the source and target details.
		// retryDeferred:
		//		Deferred retry object if either the previous create attempt fails or link creation is not allowed for link type, 
		//		or response
		
		var isRetry = retryDeferred != null;
		var remoteResourceOwnedLinkDeferred = isRetry ? retryDeferred : new dojo.Deferred();
		
		var getIsCollectionDeferred = this.isCalmSourceCollectionType(calmLink.getSubject());
		getIsCollectionDeferred.addCallback(this, function(isCollection) {
			var linkTypeDeferred = this._linkTypeURLToCalmLinkType(projectAreaUri, calmLink.getLinkType());
			linkTypeDeferred.addCallback(this, function(linkTypeSameAs) {
				var coreVersion = com.ibm.rdm.web.FRONTING_PROPERTY_ENABLE_OSLCV2_CONSUMPTION == "true" ? com.ibm.rdm.web.OSLC_Core_Version : "1.0";
				var acceptHeader = (isRetry || (coreVersion == "1.0")) ? this._getAcceptHeaderForLinkType(linkTypeSameAs, isCollection) : "application/rdf+xml";
				var getUri = calmLink.getObject();
				var deferred = RDMClient.doGet({
					uri: getUri,
					skipCache: true,
					skipErrors: !isRetry,
					xhrArgs: {
						headers: {
							"Accept": acceptHeader,
							"OSLC-Core-Version": coreVersion
						}
					},
					parser: function(response) {
						var versionReturned = response.ioArgs.xhr.getResponseHeader("OSLC-Core-Version");
						var oslcLinkUtils = null;
						if (versionReturned == "2.0") {
							oslcLinkUtils = com.ibm.rdm.web.calm.OSLCV2LinkUtils;
						} else {
							// For all other cases, assume v1.0 spec
							oslcLinkUtils = com.ibm.rdm.web.calm.OSLCV1LinkUtils;
						}
						var createDeferred = oslcLinkUtils.createBackLink(projectAreaUri, calmLink, isCollection, response, linkTypeSameAs);
						createDeferred.addCallback(this, function(retVal){
							remoteResourceOwnedLinkDeferred.callback(retVal);
						});
						createDeferred.addErrback(this, function(retVal){
							remoteResourceOwnedLinkDeferred.errback(retVal);
						});
					}
				});
				deferred.addErrback(this, function(response) {
					if (!isRetry && response.ioArgs.xhr.status != "200") {
						this._createRemoteResourceOwnedLinkDeferred(projectAreaUri, calmLink, remoteResourceOwnedLinkDeferred);
					} else {
						remoteResourceOwnedLinkDeferred.errback(response);
					}
				});
			});
			linkTypeDeferred.addErrback(this, function(response) {
				remoteResourceOwnedLinkDeferred.errback(response);
			});
		});
		return remoteResourceOwnedLinkDeferred;
	},
  
	deleteLink: function(projectAreaLink, calmLink) {
		var deleteLinkDeferred = new dojo.Deferred(); 
		
		var linkURI = calmLink.getResourceURI();
		// If we do not have a valid link URL; it must be a discovered ( from ldx ) link, just have to remote teh remote one
		if ( linkURI.indexOf("/links") > -1 ) {
			//DELETE FRONT LINK
			var deleteFrontLinkDeferred = this.deleteFrontLink(projectAreaLink, calmLink);
			deleteFrontLinkDeferred.addCallback(this, function(response) {
				if (response) {
					//DELETE BACK LINK
					var deleteBackLinkDeferred = this.deleteBackLink(projectAreaLink, calmLink);
					deleteBackLinkDeferred.addCallback(function(response) {
						deleteLinkDeferred.callback(true);
					});
					deleteBackLinkDeferred.addErrback(this, function(response){
						this._putUpWarning(7254); //cannotDeleteBacklink
						deleteLinkDeferred.callback(false);
					});
				} else {
					deleteLinkDeferred.callback(false);
				}
			});
			deleteFrontLinkDeferred.addErrback(this, function(response) {
				this._putUpError(7256); // cannotDeleteFrontlink
				deleteLinkDeferred.callback(false);
			});
		} else {
			var deleteBackLinkDeferred = this.deleteBackLink(projectAreaLink, calmLink);
			deleteBackLinkDeferred.addCallback(this, function(response) {
				// now that the backlink is gone remove the 
				// entry from the cache.
				if (calmLink.getLinkType() !== LinkTypeUtils.SYSTEM_LINK_TYPE_TRACKS_CHANGE_SET) {
    				var deleteFromCacheDeferred = this._updateLinkInCMLinkIndexer(projectAreaLink, calmLink.getSubject(), calmLink.getObject(), calmLink.getLinkType(), UPDATE_CACHE_EVENT.DELETE);
    				deleteFromCacheDeferred.then(function(response) {
    					// if updating the cache fails, it shouldn't matter - eventually the LDX will update (1 minute)
    					// Should we notify the user that it may take a few minutes to update on failure?
    					deleteLinkDeferred.callback(true);
    				});
                } else {
                    deleteLinkDeferred.callback(true);
                }
			});
			deleteBackLinkDeferred.addErrback(this, function(response){
				this._putUpWarning(7254); //cannot delete the backlink.
				deleteLinkDeferred.callback(false);
			});
		}
	  
		
		return deleteLinkDeferred;
	},
  
  /*
  *
  *This is really the backlink... (very confusing)
  *TODO: Should really refactor the name, but it might cause regressions.
  */
	deleteFrontLink:function(projectAreaLink, calmLink) {
		//Calm links are always the subject of a linktype.  The direction is needed to 
			//handle the refresh of the grid (to update the cache)	
		var direction = ArtifactConstants.SUBJECT_LINK;
		return LinkUtils.deleteLink({link: calmLink, direction: direction});
	},
  
  /*
  *
  *This is really the frontlink... (very confusing)
  *TODO: Should really refactor the name, but it might cause regressions.
  */
	deleteBackLink: function(projectAreaLink, calmLink, retryDeferred) {
		var isRetry = retryDeferred != null;
		var deleteBacklinkDeferred = isRetry ? retryDeferred : new dojo.Deferred();
		var getIsCollectionDeferred = this.isCalmSourceCollectionType(calmLink.getSubject());
		getIsCollectionDeferred.addCallback(this, function(isCollection) {
			var linkTypeDeferred = this._linkTypeURLToCalmLinkType(projectAreaLink, calmLink.getLinkType());
			linkTypeDeferred.addCallback(this, function(linkTypeSameAs) {
				var coreVersion = com.ibm.rdm.web.FRONTING_PROPERTY_ENABLE_OSLCV2_CONSUMPTION == "true" ? com.ibm.rdm.web.OSLC_Core_Version : "1.0";
				var acceptHeader = (isRetry || (coreVersion == "1.0")) ? this._getAcceptHeaderForLinkType(linkTypeSameAs, isCollection) : "application/rdf+xml";
				var deferred = RDMClient.doGet({
					uri: calmLink.getObject(),
					skipCache: true,
					skipErrors: !isRetry,
					xhrArgs: {
						headers: {
							"Accept": acceptHeader,
							"OSLC-Core-Version": coreVersion
						}
					},
					parser: function(response) {
						var versionReturned = response.ioArgs.xhr.getResponseHeader("OSLC-Core-Version");
						var oslcLinkUtils = null;
						if (versionReturned == "2.0") {
							oslcLinkUtils = com.ibm.rdm.web.calm.OSLCV2LinkUtils;
						} else {
							// For all other cases, assume v1.0 spec
							oslcLinkUtils = com.ibm.rdm.web.calm.OSLCV1LinkUtils;
						}
						var deleteDeferred = oslcLinkUtils.deleteBackLink(projectAreaLink, calmLink, isCollection, response, linkTypeSameAs);
						deleteDeferred.addCallback(this, function(retVal){
							deleteBacklinkDeferred.callback(retVal);
						});
						deleteDeferred.addErrback(this, function(retVal){
							deleteBacklinkDeferred.errback(retVal);
						});
					}
				});
				deferred.addErrback(this, function(response) {
					if (!isRetry && response.ioArgs.xhr.status != "200") {
						this.deleteBackLink(projectAreaLink, calmLink, deleteBacklinkDeferred);
					} else {
						deleteBacklinkDeferred.errback(response);
					}
				});
			});
			linkTypeDeferred.addErrback(this, function(response) {
				deleteBacklinkDeferred.errback(response);
			});
		});
		return deleteBacklinkDeferred;
	},
	
	deleteBacklinks: function(projectURI, resourceURI) {
		var linkDeleteDeferred = new dojo.Deferred();
		var linksDeferred = LinkUtils.getLinksForResource({
			sourceAndTargetURI: resourceURI,
			skipCache: true
		});
		linksDeferred.addCallback(this, function(args) {
			var links = args.resourceInfos;
			var linkTypesDeferred = LinkTypeUtils.getLinkTypesForProject(projectURI);
			linkTypesDeferred.addCallback(this, function(linkTypes) {
				var deleteBacklinkDeferreds = [];
				for (var i = 0; i < links.length; i++) {
					var link = links[i];
					for (var x = 0; x < linkTypes.length; x++) {
						if (link.getLinkType() === linkTypes[x].getResourceURI()) {
							if (LinkTypeUtils.isCalmLink(linkTypes[x].getSameAs())) {
								deleteBacklinkDeferreds.push(this.deleteBackLink(projectURI, link));
							}
							break;
						}
					}
				}
				var deleteBacklinksDeferred = new dojo.DeferredList(deleteBacklinkDeferreds);
				deleteBacklinksDeferred.addBoth(this, function(deleteBacklinksResponse) {
					linkDeleteDeferred.callback(deleteBacklinksResponse);
				});
			});
			linkTypesDeferred.addErrback(this, function(linkTypesFailure) {
				linkDeleteDeferred.errback(linkTypesFailure);
			});
		});
		linksDeferred.addErrback(this, function(linkFailure) {
			linkDeleteDeferred.errback(linkFailure);
		});
		return linkDeleteDeferred;
	},
  
	getOldProjectLinkTypeFromSameAs: function(sameAs) {
		var retVal = null;
		switch (sameAs) {
			case LinkTypeUtils.SYSTEM_LINK_TYPE_IMPLEMENTED_BY:
				retVal = CalmConstants.ProjectAreaLinkType.implementedBy;
				break;
			case LinkTypeUtils.SYSTEM_LINK_TYPE_TRACKED_BY:
				retVal = CalmConstants.ProjectAreaLinkType.trackedBy;
				break;
			case LinkTypeUtils.SYSTEM_LINK_TYPE_VALIDATED_BY:
				retVal = CalmConstants.ProjectAreaLinkType.validatedBy;
				break;
		}
		return retVal;
	},
	
	isCalmSourceCollectionType: function(resourceURI) {
		//Returns true/false is resourceURI is a collection type
		var fakeDeferred = new dojo.Deferred();
		
		if (com.ibm.rdm.web.query.SavedFilterUtils.isSavedView(resourceURI)) {
			fakeDeferred.callback(true);
		} else {
			var descriptionsDeferred = com.ibm.rdm.web.model.resources.ResourceManager.getResourceDescribe({resourceURI: resourceURI});
			descriptionsDeferred.addCallback(this, function(response){	
			    var isCollection = false;
			    if (response) {
    				var artifactType = response.contentType;
    				var projectURI = response.projectURI;
    				isCollection = artifactType == MimeTypes.getArtifactFormat(MimeTypes.COLLECTION_NAME, projectURI)
    					|| artifactType == MimeTypes.getArtifactFormat(MimeTypes.MODULE_NAME, projectURI);
    			}
								
				fakeDeferred.callback(isCollection);
				
			});
			descriptionsDeferred.addErrback(this, function(response){	
				  console.error("Error retrieving resource description");
				  fakeDeferred.callback(false);
			});
		}
		
		return fakeDeferred;
	}, 
	
	_putUpError: function(errorNumber){
		com.ibm.rdm.web.common.MessageHandler.showError(errorNumber);
	},
	
	_putUpWarning: function(warningNumber){
		com.ibm.rdm.web.common.MessageHandler.showWarning(warningNumber);
	},
	
	getProjectLinkTypeDisplayLabel: function(linkType){
		var linkTypeLabel = null;
		switch (linkType) {
			case CalmConstants.ProjectAreaLinkType.implementedBy:
				linkTypeLabel = calmMessages.implementedBy;
				break;
			case CalmConstants.ProjectAreaLinkType.trackedBy:
				linkTypeLabel = calmMessages.trackedBy;
				break;
			case CalmConstants.ProjectAreaLinkType.validatedBy:
				linkTypeLabel = calmMessages.validatedBy;
				break;
			case CalmConstants.ProjectAreaLinkType.elaboratedBy:
				//Elaborated Types are used by DM links,
				linkTypeLabel = calmMessages.derivedTypeLink;
				break;
			case CalmConstants.ProjectAreaLinkType.relatedTo:
				//Link from RM to DWA
				linkTypeLabel = calmMessages.referencesLink;
				break;
 			case CalmConstants.ProjectAreaLinkType.refine:
			    linkTypeLabel = calmMessages.refineLink;
				break;	
			case CalmConstants.ProjectAreaLinkType.trace:
			    linkTypeLabel = calmMessages.traceLink;
				break;							
			case CalmConstants.ProjectAreaLinkType.satisfy_dm:
				linkTypeLabel = calmMessages.satisfyLink_dm;
				break;				
			default:
				break;
		}
		return linkTypeLabel;
	}

};
})();
